#lang racket/base
(require racket/match
         "global.rkt"
         "nums.rkt")

(provide disassemble
         assemble)

;; Assembling and disassembling symbolic instructions

;; disassemble : num -> (listof (union num sym))
;; note: produces a symbolic instruction from number
(define (disassemble num)
  (let ([first  (extract num 0 1)]
	[second (extract num 1 1)]
	[third  (extract num 2 1)]
	[R (lambda (x) (string->symbol (format "R~a" x)))])
    (cond
      ((not (zero? first)) ; I format instructions
       (let ((x (R (extract num 1 1)))
	     (i (extract num 2 6)))
	 (case first
	   ((1)    `(jmpi ,i))
	   ((2)    `(jsr ,x ,i))
	   ((3)    `(bez ,x ,i))
	   ((4)    `(bnez ,x ,i))
	   ((5)    `(blz ,x ,i))
	   ((6)    `(blez ,x ,i))
	   ((7)    `(bgz ,x ,i))
	   ((8)    `(bgez ,x ,i))
	   ((9)    `(ldi ,x ,(if (negative? num) (- 0 i) i)))
	   (else '(unknown)))))
      ((not (zero? second)) ; R format instruction
       (let ([x (R (extract num 2 1))]
	     [y (R (extract num 3 1))]
	     [z (R (extract num 4 1))]
	     [d (extract num 5 1)])
	 (case second
	   ((1)    `(add ,x ,y ,z))
	   ((2)    `(sub ,x ,y ,z))
	   ((3)    `(mul ,x ,y ,z))
	   ((4)    `(div ,x ,y ,z))
	   ((5)    `(mod ,x ,y ,z))
	   ((6)    (if (zero? d)
		       `(ldx ,x ,y ,z)
		       `(ld ,x ,y)))
	   ((7)    (if (zero? d)
		       `(stx ,x ,y ,z)
		       `(st ,x ,y)))
	   ((8)    `(jmpx ,x))
	   ((9)    `(mov ,x ,y))
	   (else '(unknown)))))
      (else ; S format instruction
	(let ((x (R (extract num 3 1))))
	  (case third
	    ((0)    `(halt))
	    ((1)    `(print ,x))
	    ((2)    `(newline))
	    (else '(unknown))))))))

; assemble : (listof (union sym num)) -> num
;; note: produces a number from symbolic instruction
(define (assemble inst)
  ;; encoding the instructions
  (define (jmpi i)       (encode  i `(1   0 ,i)))
  (define (jsr Rx i)     (encode  i `(2 ,Rx ,i)))
  (define (bez Rx i)     (encode  i `(3 ,Rx ,i)))
  (define (bnez Rx i)    (encode  i `(4 ,Rx ,i)))
  (define (blz Rx i)     (encode  i `(5 ,Rx ,i)))
  (define (blez Rx i)    (encode  i `(6 ,Rx ,i)))
  (define (bgz Rx i)     (encode  i `(7 ,Rx ,i)))
  (define (bgez Rx i)    (encode  i `(8 ,Rx ,i)))
  (define (ldi Rx i)     (encode  i `(9 ,Rx ,i)))
  
  (define (add Rx Ry Rz) (encode +1 `(0 1 ,Rx ,Ry ,Rz)))
  (define (sub Rx Ry Rz) (encode +1 `(0 2 ,Rx ,Ry ,Rz)))
  (define (mul Rx Ry Rz) (encode +1 `(0 3 ,Rx ,Ry ,Rz)))
  (define (div Rx Ry Rz) (encode +1 `(0 4 ,Rx ,Ry ,Rz)))
  (define (mod Rx Ry Rz) (encode +1 `(0 5 ,Rx ,Ry ,Rz)))
  (define (ld Rx Ry)     (encode +1 `(0 6 ,Rx ,Ry   0 1)))
  (define (ldx Rx Ry Rz) (encode +1 `(0 6 ,Rx ,Ry ,Rz 0)))
  (define (st Rx Ry)     (encode +1 `(0 7 ,Rx ,Ry   0 1)))
  (define (stx Rx Ry Rz) (encode +1 `(0 7 ,Rx ,Ry ,Rz 0)))
  (define (jmpx Rx)      (encode +1 `(0 8 ,Rx)))
  (define (mov Rx Ry)    (encode +1 `(0 9 ,Rx ,Ry)))

  (define (halt)         (encode +1 `(0 0 0   0)))
  (define (print Rx)     (encode +1 `(0 0 1 ,Rx)))
  (define (newline)      (encode +1 `(0 0 2   0)))

  (define (encode s lop)
    (* (sign s) (foldr (lambda (x y) (+ (abs (register> x)) (* 10 y))) 0 lop)))
  
  (match inst
    [(list 'add (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (add Rx Ry Rz)]
    [(list 'sub (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (sub Rx Ry Rz)]
    [(list 'mul (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (mul Rx Ry Rz)]
    [(list 'div (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (div Rx Ry Rz)]
    [(list 'mod (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (mod Rx Ry Rz)]
    [(list 'ldi (? symbolic-register? Rx) (? immediate? i))
     (ldi Rx i)]
    [(list 'ld  (? symbolic-register? Rx) (? symbolic-register? Ry))
     (ld Rx Ry)]
    [(list 'ldx (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (ldx Rx Ry Rz)]
    [(list 'st  (? symbolic-register? Rx) (? symbolic-register? Ry))
     (st Rx Ry)]
    [(list 'stx (? symbolic-register? Rx) (? symbolic-register? Ry) (? symbolic-register? Rz))
     (stx Rx Ry Rz)]
    [(list 'mov (? symbolic-register? Rx) (? symbolic-register? Ry))
     (mov Rx Ry)]
    [(list 'jmpx (? symbolic-register? Rx))
     (jmpx Rx)]
    [(list 'jmpi (? immediate? i))
     (jmpi i)]
    [(list 'jsr (? symbolic-register? Rx) (? immediate? i))
     (jsr Rx i)]
    [(list 'bez (? symbolic-register? Rx) (? immediate? i))
     (bez Rx i)]
    [(list 'bnez (? symbolic-register? Rx) (? immediate? i))
     (bnez Rx i)]
    [(list 'blz (? symbolic-register? Rx) (? immediate? i))
     (blz Rx i)]
    [(list 'blez (? symbolic-register? Rx) (? immediate? i))
     (blez Rx i)]
    [(list 'bgz (? symbolic-register? Rx) (? immediate? i))
     (bgz Rx i)]
    [(list 'bgez (? symbolic-register? Rx) (? immediate? i))
     (bgez Rx i)]
    [(list 'halt)
     (halt)]
    [(list 'print (? symbolic-register? Rx))
     (print Rx)]
    [(list 'newline)
     (newline)]
    [(list 'data (? immediate? d))
     d]
    [else (error 'assemble "bad instruction format: ~s" inst)]))

;; immediate? : any -> bool
(define (immediate? n) 
  (and (exact-integer? n) 
       (< (abs n) MAX-IMM)))

;; symbolic-register? : any -> bool
(define (symbolic-register? r)
  (or (and (exact-integer? r)
           (<= 0 r MAX-REG-INDEX))
      (and (symbol? r)
           (let* ((r (symbol->string r))
                  (d (string->number (substring r 1 (string-length r)))))
             (and d
                  (eq? #\R (string-ref r 0))
                  (<= 0 d MAX-REG-INDEX))))))

;; register->index : {'R0,...,'R9} -> num[(<= 0 d MAX-REG-INDEX)]
;; Warning: no domain check!
(define (register->index r)
  (let ((r (symbol->string r)))
    (string->number (substring r 1 (string-length r)))))

;; register> : (union sym num) -> num
(define (register> r)
  (if (symbol? r)
      (register->index r)
      r))
